EnableExplicit

;{ CONSTANTS 

Enumeration  
 ; RETURN VALUES
 #SORT_STRUCT_OK = 0
 #SORT_STRUCT_ERR_INVALID_PARAMETERS 
EndEnumeration

;}

;{ STRUCTURES

Structure T_PBLIST_HDR 
; DESC
;  Each items in the linked list have this structure as header, followed by the actual data.
;  This one is documented, so it should be safe to use it

 *NextItem.T_PBLIST_HDR ; Pointer to next element in list or zero if at last element
 *PrevItem.T_PBLIST_HDR ; Pointer to previous element in list or zero if at first element
EndStructure

;}

;{ DECLARES

Declare.i 	SortStructuredArrayEx (*first, *last, iNumElements, iSize, *fpCompare)
Declare.i 	SortStructuredListEx (*first, *last, iNumElements, iSize, *fpCompare, flgPbList = 1)

;}

Macro PBLIST_HDR_ADD (ptr_list_item)
; DESC
;  Calculate the address of the PB linked list header associated with the passed
;  linked list item pointer.
;
; IN
; 	ptr_listitem    ; Address of a linked list item.
;
; OUT
;  The address of associated the T_PBLIST_HDR structure.
;
; EXAMPLE
;  *ListHeader = PBLIST_HDR_ADD(FirstElement(lsList()))

 (ptr_list_item - SizeOf(T_PBLIST_HDR))
EndMacro

Procedure.i SortStructuredArrayEx (*first, *last, iNumElements, iSize, *fpCompare) 

; DESC
;  Sort a PB structured array using an external compare procedure. 
;  This allows to sort using multiple fields or implement other custom sort algorithms.
;
; IN
;  *first		    ; Pointer to the first element of the array
;  *last 	        ; Pointer to the last element of the array
;  iNumElements    ; Number of elements in the array
;  iSize	        ; Size in bytes of the data contained in the associated structure
;  *fpCompare      ; Function pointer to the compare procedure
;
; OUT
;  The sorted array.
; 
; RET
;  #PBL_OK
;  #SORT_STRUCT_ERR_INVALID_PARAMETERS  
;
; NOTES
;	This procedure is derived from the Quick Sort for linked lists written by Horst Schaeffer.
;  This procedure use recursion.
;  See the helper macro SortStructuredArrayExM.
;
; EXAMPLE
;	iRetVal = SortStructuredArrayEx (@Array(0), @Array(99), 100, iSize, @Compare())
  
 Protected *lower   = *first  
 Protected *upper   = *last 
 Protected *item    = *first  
 Protected TempBuffer$ = Space(iSize)  
  
 Protected flgGap, lCountLow, lCountHi
 
 
 If *first = #Null Or *last = #Null Or *fpCompare = #Null
    ProcedureReturn #SORT_STRUCT_ERR_INVALID_PARAMETERS
 EndIf

 If iNumElements <= 0 Or iSize <= 0
    ProcedureReturn #SORT_STRUCT_ERR_INVALID_PARAMETERS
 EndIf

  
 *item + (iNumElements / 2) * iSize
 
 CopyMemory (*item, @TempBuffer$, iSize)          
 
 CopyMemory (*first, *item, iSize)         
  
 While *lower <> *upper 
    If flgGap 
        If CallFunctionFast (*fpCompare, *lower, @TempBuffer$) = #False 
            CopyMemory(*lower, *upper, iSize) : flgGap = #False
        EndIf 
    Else 
        If CallFunctionFast (*fpCompare, @TempBuffer$, *upper) = #False
            CopyMemory (*upper, *lower, iSize) : flgGap = #True
        EndIf 
    EndIf 

    If flgGap 
        *lower + iSize
        lCountLow + 1 
    Else
        *upper - iSize
        lCountHi + 1 
    EndIf 
 Wend 
 
 CopyMemory(@TempBuffer$, *lower, iSize)  
 
 If lCountLow > 1 
    *lower - iSize
    SortStructuredArrayEx (*first, *lower, lCountLow, iSize, *fpCompare) 
 EndIf 
 
 If lCountHi > 1     
    *upper + iSize
    SortStructuredArrayEx (*upper, *last, lCountHi, iSize, *fpCompare) 
 EndIf   
  
 ProcedureReturn #SORT_STRUCT_OK
EndProcedure 

Macro SortStructuredArrayExM (arrayname, size, fpcompare)
; DESC
;  Sort a PB structured array using an external compare procedure. 
;  This allows to sort using multiple fields or implement other custom sort algorithms.
;  This is an helper macro to simplify the call to SortStructuredArrayEx().
;
; IN
;  arrayname    ; Name of the array, without @ or ()
;  size	        ; Size in bytes of the data contained in the associated structure.
;  *fpCompare   ; Function pointer to the compare procedure.
;
; OUT
;  The sorted array.
;
; RET
;  #PBL_OK
;  #SORT_STRUCT_ERR_INVALID_PARAMETERS  
;
; NOTES
;  The called procedure use recursion.
;
; EXAMPLE
;	iRetVal = SortStructuredArrayExM (Array, SizeOf(T_ITEM_STRUCT), @Compare())

 SortStructuredArrayEx (@arrayname(0), @arrayname(ArraySize(arrayname())), ArraySize(arrayname()) - 1, size, fpcompare) 
EndMacro


Procedure.i SortStructuredListEx (*first, *last, iNumElements, iSize, *fpCompare, flgPbList = 1)

; DESC
;  Sort a PB structured linked list using an external compare procedure. 
;  This allows to sort using multiple fields or implement other custom sort algorithms.
;
; IN
;  *first		    ; Pointer to the header of the first element of the list.
;  *last 	        ; Pointer to the header of the last element of the list.
;  iNumElements     ; Number of elements in the list.
;  iSize	        ; Size in bytes of the data item + the header size.
;  *fpCompare       ; Function pointer to the compare procedure.
;  flgPbList        ; = 1 if the passed list is a PureBasic native linked list (default)
;                   ; = 0 if the passed list is a LinkedListEx linked list
;
; OUT
;  The sorted linked list.
; 
; RET
;  #PBL_OK
;  #SORT_STRUCT_ERR_INVALID_PARAMETERS  
;
; NOTES
;	This procedure is derived from the Quick Sort for linked lists written by Horst Schaeffer.
;  This procedure use recursion.
;  See the helper macro SortStructuredListExM.
;
; EXAMPLE
;	iRetVal = SortStructuredListEx (PBLIST_HDR_ADD(FirstElement(lsList())), PBLIST_HDR_ADD(LastElement(lsList())), ListSize(lsList()), SizeOf(T_PBLIST_HDR) + iSize, @Compare(), 1)

 Protected *lower.T_PBLIST_HDR    = *first  
 Protected *upper.T_PBLIST_HDR    = *last 
 Protected *item.T_PBLIST_HDR     = *first  
 Protected iToGo = iNumElements / 2
 Protected iHeaderSize = SizeOf(T_PBLIST_HDR) 
 Protected flgGap, iCountLow, iCountHi
 Protected TempBuffer$
 

 If *first = #Null Or *last = #Null Or *fpCompare = #Null
    ProcedureReturn #SORT_STRUCT_ERR_INVALID_PARAMETERS
 EndIf

 If iNumElements <= 0 Or iSize <= 0
    ProcedureReturn #SORT_STRUCT_ERR_INVALID_PARAMETERS
 EndIf

 If flgPbList
    TempBuffer$ = Space(iSize - iHeaderSize)
 Else
    TempBuffer$ = Space(iSize) 
 EndIf
     
 
 Repeat 
    *item = *item\NextItem : iToGo - 1 
 Until iToGo = 0


 If flgPbList
    CopyMemory (*item + iHeaderSize, @TempBuffer$, iSize - iHeaderSize)
 Else
    CopyMemory (*item, @TempBuffer$, iSize)              
 EndIf
 
 
 CopyMemory (*first + iHeaderSize, *item + iHeaderSize, iSize - iHeaderSize)
 
  
 While *lower <> *upper ;And *upper <> 0
    If flgGap 
        If flgPbList
            If CallFunctionFast (*fpCompare, *lower + iHeaderSize, @TempBuffer$) = #False 
                CopyMemory(*lower + iHeaderSize, *upper + iHeaderSize, iSize - iHeaderSize) : flgGap = #False
            EndIf 
        Else
            If CallFunctionFast (*fpCompare, *lower, @TempBuffer$) = #False 
                CopyMemory(*lower + iHeaderSize, *upper + iHeaderSize, iSize - iHeaderSize) : flgGap = #False
            EndIf 
        EndIf
    Else 
        If flgPbList
            If CallFunctionFast (*fpCompare, @TempBuffer$, *upper + iHeaderSize) = #False
                CopyMemory (*upper + iHeaderSize, *lower + iHeaderSize, iSize - iHeaderSize) : flgGap = #True
            EndIf
        Else
            If CallFunctionFast (*fpCompare, @TempBuffer$, *upper) = #False
                CopyMemory (*upper + iHeaderSize, *lower + iHeaderSize, iSize - iHeaderSize) : flgGap = #True
            EndIf
        EndIf        
    EndIf 
    
    If flgGap 
        *lower = *lower\NextItem
        iCountLow + 1 
    Else
        *upper = *upper\PrevItem
        iCountHi + 1 
    EndIf 
 Wend 
 
 If flgPbList
    CopyMemory(@TempBuffer$, *lower + iHeaderSize, iSize - iHeaderSize)  
 Else
    CopyMemory(@TempBuffer$ + iHeaderSize, *lower + iHeaderSize, iSize - iHeaderSize)  
 EndIf
 
 
 If iCountLow > 1     
    *lower = *lower\PrevItem
    SortStructuredListEx (*first, *lower, iCountLow, iSize, *fpCompare, flgPbList)     
 EndIf 
 
 If iCountHi > 1     
    *upper = *upper\NextItem
    SortStructuredListEx (*upper, *last, iCountHi, iSize, *fpCompare, flgPbList)     
 EndIf 
 
 ProcedureReturn #SORT_STRUCT_OK  
EndProcedure 

Macro SortStructuredListExM (listname, size, fpcompare)

; DESC
;  Sort a PB linked list using an external compare procedure. 
;  This allows to sort using multiple fields or implement other custom sort algorithms.
;  This is an helper macro to simplify the call to SortStructuredListEx().
;
; IN
; 	listname    ; Name of the list *without* the ()
;  size        ; Size in bytes of the data item contained in the list.
;  fpcompare   ; Address of the compare procedure.
;
; OUT
;  The sorted linked list.
;
; RET
;  #PBL_OK
;  #SORT_STRUCT_ERR_INVALID_PARAMETERS  
;
; NOTES
;  The called procedure use recursion.
;
; EXAMPLE
;  iRetVal = SortStructuredListExM (lsList, SizeOf(T_ITEM_STRUCT), @Compare())
;
; HISTORY
;  Sep 16, 2008    - In PB 4.30 the linked list procedures were modified to return a pointer to the actual
;                    data instead of the list header. The use of the PBLIST_HDR_ADD addressed the problem.

 SortStructuredListEx (PBLIST_HDR_ADD(FirstElement(listname())), PBLIST_HDR_ADD(LastElement(listname())), ListSize(listname()), SizeOf(T_PBLIST_HDR) + size, fpcompare, 1)
EndMacro

; IDE Options = PureBasic 5.73 LTS (Windows - x86)
; CursorPosition = 17
; Folding = --
; EnableXP
; EnableUser
; Executable = ..\out32\i2pdf.exe
; CPU = 1
; CompileSourceDirectory
; EnableBuildCount = 1